from overrides import overrides
from typing import List
import threading

from neovi.neodevice import NeoDevice
from neovi.neovi import SUCCESS
from neovi.structures import icsSpyMessage


class OfflineNeoDevice(NeoDevice):
    def __init__(self, messages: List[icsSpyMessage]) -> None:
        super().__init__(None, False, False)
        self.messages = messages

    @overrides
    def get_messages(self):
        return (SUCCESS, self.messages, [])


class TestSubscriptions:
    def do_subscription_test(self, messages: List[icsSpyMessage], expected_output: List[icsSpyMessage], network_filter=None, address_filter=None, msg_type_filter=None, additional_bytes_filter=None) -> None:
        remaining = expected_output[:]
        unexpected = []

        def callback(msg, user_data):
            if msg in remaining:
                remaining.remove(msg)
            else:
                unexpected.append(msg)

        neodevice = OfflineNeoDevice(messages)
        neodevice.subscribe_to(callback, network_filter, address_filter, msg_type_filter, additional_bytes_filter)
        neodevice._process_msg_queue()
        assert len(remaining) == 0
        assert len(unexpected) == 0


    def test_subscription(self) -> None:
        messages = [icsSpyMessage()]
        self.do_subscription_test(messages, messages)


    def test_subscription_network_filter(self) -> None:
        correct_network = icsSpyMessage(NetworkID=1)
        incorrect_network = icsSpyMessage(NetworkID=2)
        self.do_subscription_test([correct_network, incorrect_network], [correct_network], network_filter=1)


    def test_subscription_address_filter(self) -> None:
        correct_address = icsSpyMessage(ArbIDOrHeader=1)
        incorrect_address = icsSpyMessage(ArbIDOrHeader=2)
        self.do_subscription_test([correct_address, incorrect_address], [correct_address], address_filter=1)


    def test_subscription_msg_type_filter(self) -> None:
        correct_msg_type = icsSpyMessage()
        correct_msg_type.Data[1] = 1
        incorrect_msg_type = icsSpyMessage()
        incorrect_msg_type.Data[1] = 2
        self.do_subscription_test([correct_msg_type, incorrect_msg_type], [correct_msg_type], msg_type_filter=1)


    def test_subscription_additional_bytes_filter(self) -> None:
        correct_additional_bytes = icsSpyMessage()
        correct_additional_bytes.Data[2] = 2
        correct_additional_bytes.Data[3] = 5
        incorrect_additional_bytes = icsSpyMessage()
        incorrect_additional_bytes.Data[2] = 1
        incorrect_additional_bytes.Data[3] = 2
        self.do_subscription_test([correct_additional_bytes, incorrect_additional_bytes], [correct_additional_bytes], additional_bytes_filter=[2, 5])


    def test_subscription_auto_remove(self) -> None:
        def callback(msg, user_data):
            pass

        neodevice = OfflineNeoDevice([icsSpyMessage()])

        neodevice.subscribe_to(callback, auto_remove=True)
        neodevice._process_msg_queue()
        assert len(neodevice._subscriptions) == 0

        neodevice.subscribe_to(callback, auto_remove=False)
        neodevice._process_msg_queue()
        assert len(neodevice._subscriptions) == 1


    def test_subscription_user_data(self) -> None:
        submitted_user_data = "test string"
        success = threading.Event()

        def callback(msg, user_data):
            if submitted_user_data == user_data:
                success.set()

        neodevice = OfflineNeoDevice([icsSpyMessage()])
        neodevice.subscribe_to(callback, user_data=submitted_user_data)
        neodevice._process_msg_queue()
        assert success.is_set()
